/**
 * This program demonstrates volumetric rendering in OpenGL using 2d
 * or 3d textures.
 *
 * Keys:
 * '2' - use 2d textures
 * '3' - use 3d textures
 * 'a' - animate, i.e. simulate the field
 * 'r' - reset the field (to the initial torus)
 *
 * Use the mouse to rotate the volume and the arrow keys to translate it.
 *
 * Click (without dragging) to create new tori within the field. The pick
 * ray will be shown in red, its intersection with the volume in green.
 * (Note that by definition, you can't immediately see a pick ray until
 * you rotate the view, since it lies along the view vector.)
 *
 * The volumetric functions are fieldTexture() and display().
 * The pick ray is handled in mouse() and segmentCubeIntersect().
 *
 * Harold Cooper <harold@mit.edu> 2006-01-22
 **/

//#include <sys/time.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <GLUT/glut.h>
#include "field.h"

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif
#define DEG2RAD (M_PI/180)


float thetax=0, thetay=0, dx=0, dz=0, // viewpoint
  omega=1, velocity=0.1; // rate of user motion
grid field; // the field
char texture3D = FALSE; // 3d texturing? (as opposed to 2d)
GLuint *tex2D; // array of 2d textures

char animating = FALSE; // used by idle()
char rotating = FALSE; int rotatex0, rotatey0; // used by mouse() and motion()
char pickCube = FALSE;
double near[3]={0,0,0}, far[3]={0,0,0}, pick0[3], pick1[3]; // the pick ray


/*** If there's a GL error, print msg and then the error. ***/
int printGLError(char *msg) {
  GLenum err = glGetError();
  const GLubyte *str;
  if(err) {
    printf("at \"%s\":\n",msg);
    str = gluErrorString(err);
    if(str) printf("GL error %d: %s\n", (int)err, str);
    else printf("GL error %d: unknown error\n", (int)err);
  }
  return err;
}

/*** Upload the data in field to the current texture(s). ***/
void fieldTexture(void) {
  int i, plane = field.dim*field.w*field.h;
  GLenum format = (field.dim==3)?GL_RGB:GL_LUMINANCE; // could also consider dim==4 as GL_RGBA, for alpha blending
  glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
  
  if(texture3D) { // a single 3d texture
    //glBindTexture( GL_TEXTURE_3D, tex3D );
    glTexParameterf(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    glTexParameterf(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    glTexParameterf(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP );
    glTexParameterf(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP );
    glTexParameterf(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP );
    glTexImage3D( GL_TEXTURE_3D, 0, format, field.w, field.h, field.d, 0, format, GL_FLOAT, field.v );
    //gluBuild3DMipmaps( GL_TEXTURE_3D, format, field.w, field.h, field.d, format, GL_FLOAT, field.v );
  }
  else { // several 2d textures
    // for 2d, we need a separate texture for each slice
    for(i=0;i<field.d;i++) {
      glBindTexture(GL_TEXTURE_2D, tex2D[i]);
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP );
      glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP );
      glTexImage2D( GL_TEXTURE_2D, 0, format, field.w, field.h, 0, format, GL_FLOAT, field.v+i*plane);
    }
  }
  printGLError("fieldTexture");
}

/*** Sets up GL environment and field. ***/
void init(void) {
  //glClearColor(0.25,0.25,0.25,1);
  glEnable(GL_DEPTH_TEST);
  //glPolygonMode(GL_BACK, GL_LINE);
  
  grid_alloc(&field, 32, 32, 32, 3);
  initField(&field);
  
  tex2D = (GLuint *)malloc(sizeof(GLuint)*field.d);
  glGenTextures(field.d,tex2D);
  printGLError("init");
  fieldTexture();
}

/*** glutReshapeFunc ***/
void reshape(int width, int height) {
  glViewport(0, 0, width, height);
  float aspectRatio = (height==0)?width:width/(float)height;

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(60, aspectRatio, 1, 30);
  
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glTranslatef(0,0,-2);
}

/*** transform the current matrix for display ***/
void displayMatrixTransform(void) {
    glTranslatef(dx,0,dz);
    glRotatef(thetax,1,0,0);
    glRotatef(thetay,0,1,0);
    glTranslatef(-0.5,-0.5,-0.5); // so (0,0) and (1,1) are symmetric about origin
}

/*** glutDisplayFunc ***/
void display(void) {
  int i;
  float t,
    x = cos(thetax*DEG2RAD)*sin(thetay*DEG2RAD),
    y = -sin(thetax*DEG2RAD),
    z = -cos(thetax*DEG2RAD)*cos(thetay*DEG2RAD); // (x,y,z) is direction of view
  
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
    displayMatrixTransform();
    
    /*** draw the pick rays, recorded by mouse(): ***/
    //glLineWidth(2);
    glBegin(GL_LINES);
    glColor3f(1,0,0);
    glVertex3dv(near);
    glVertex3dv(far);
    glEnd();
    if(pickCube) {
      glClear(GL_DEPTH_BUFFER_BIT);
      glDisable(GL_BLEND);
      glBegin(GL_LINES);
      glColor3f(0,1,0);
      glVertex3dv(pick0);
      glVertex3dv(pick1);
      glEnd();
    }
    
    glColor3f(1,1,1); // reset color
    
    glEnable(GL_BLEND);
    //glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
    glBlendFunc( GL_SRC_COLOR, GL_ONE_MINUS_SRC_COLOR ); // use luminance as alpha

    if(texture3D) { // a single 3d texture
      glEnable(GL_TEXTURE_3D);
      glBegin(GL_QUADS);
      
      /*** Draw slices as close to perpendicular to viewpoint as possible,
        while still aligned with axes: ***/
        if(fabs(x)>=fabs(y)&&fabs(x)>=fabs(z)) { // viewpoint primarily along x-axis
          for(i=0; i<field.w; i++) { // draw slices perpendicular to x-axis
            t = i/(float)field.w;
            if(x>0) t = 1-t; // always draw from back to front, for blending to work
            glTexCoord3f(t,0,0); glVertex3f(t,0,0);
            glTexCoord3f(t,1,0); glVertex3f(t,1,0);
            glTexCoord3f(t,1,1); glVertex3f(t,1,1);
            glTexCoord3f(t,0,1); glVertex3f(t,0,1);
          }
        }
        else if(fabs(y)>=fabs(z)) { // viewpoint primarily along y-axis
          for(i=0; i<field.h; i++) {
            t = i/(float)field.h;
            if(y>0) t = 1-t;
            glTexCoord3f(0,t,0); glVertex3f(0,t,0);
            glTexCoord3f(1,t,0); glVertex3f(1,t,0);
            glTexCoord3f(1,t,1); glVertex3f(1,t,1);
            glTexCoord3f(0,t,1); glVertex3f(0,t,1);
          }
        }
        else { // viewpoint primarily along z-axis
          for(i=0; i<field.d; i++) {
            t = i/(float)field.d;
            if(z>0) t = 1-t;
            glTexCoord3f(0,0,t); glVertex3f(0,0,t);
            glTexCoord3f(1,0,t); glVertex3f(1,0,t);
            glTexCoord3f(1,1,t); glVertex3f(1,1,t);
            glTexCoord3f(0,1,t); glVertex3f(0,1,t);
          }
        }

      glEnd();
      glDisable(GL_TEXTURE_3D);
    }
    else { // several 2d textures
      /*** in 2d, we can only slice along the z-axis, since we're storing the
        field data and thus the textures in z-major order; if we tried to slice
        along another axis, the indexing would be crazy, i.e. really slow. ***/
      for(i=0; i<field.d; i++) { // slice along z-axis, regardless of viewpoint
        if(z>0) i=field.d-1-i; // always draw from back to front
        glBindTexture(GL_TEXTURE_2D, tex2D[i]); // each slice has its own texture
        glEnable(GL_TEXTURE_2D);
        glBegin(GL_QUADS);
        t = i/(float)field.d;
        glTexCoord2f(0,0); glVertex3f(0,0,t);
        glTexCoord2f(1,0); glVertex3f(1,0,t);
        glTexCoord2f(1,1); glVertex3f(1,1,t);
        glTexCoord2f(0,1); glVertex3f(0,1,t);
        if(z>0) i=field.d-1-i;
        glEnd();
        glDisable(GL_TEXTURE_2D);
      }
    }

  glPopMatrix();

  //glFlush();
  glutSwapBuffers();
  printGLError("display");
}

/*** glutIdleFunc ***/
void idle(void) {
  if(animating) {
    stepField(&field);
    fieldTexture();
    glutPostRedisplay();
  }
}

/*** glutKeyboardFunc ***/
void keyboard(unsigned char key, int x, int y) {
  switch(key) {
    case '2':
      if(texture3D) {
        texture3D = FALSE;
        fieldTexture();
        glutPostRedisplay();
      }
      break;
    case '3':
      if(!texture3D) {
        texture3D = TRUE;
        fieldTexture();
        glutPostRedisplay();
      }
      break;
    case 'a':
      animating = !animating;
      break;
    case 'r': // 'reset'
      initField(&field);
      fieldTexture();
      glutPostRedisplay();
      break;
  }
}

/*** glutSpecialFunc ***/
void special(int key, int x, int y) {
  switch(key) {
    case GLUT_KEY_UP:
      dz-=velocity;
      break;
    case GLUT_KEY_DOWN:
      dz+=velocity;
      break;
    case GLUT_KEY_LEFT:
      dx-=velocity;
      break;
    case GLUT_KEY_RIGHT:
      dx+=velocity;
      break;
  }
  glutPostRedisplay();
}

/** Returns TRUE if the segment between a0 and b0 intersects the unit cube,
 * and records the two points of intersection in a1 and b1. If the intersection
 * is a line segment, i.e. is contained in a face, then a1 and b1 are the
 * endpoints of the line segment.
 * Returns FALSE if the segment does not intersect the cube.
 **/
char segmentCubeIntersect(double *a0, double *a1, double *b0, double *b1) {
  static const double eps = 0.001; // intersection tolerance
  double s, t, *seg[2]={b0,b1};
  int n=0, i, j, w;
  
  for(i=0;i<6&&n<2;i++) { // 6 faces of the cube
    j = i/2; // axis
    w = i%2; // coord on axis
    if(abs(a1[j]-a0[j])>=eps) {
      t = (w-a0[j]) / (a1[j]-a0[j]);
      if(-eps<t && t<1+eps) {
        s = 1-t;
        for(j=0;j<3;j++) {
          seg[n][j] = a0[j]*s+a1[j]*t;
          if(seg[n][j]<=-eps || seg[n][j]>=1+eps)
            break;
        }
        if(j==3) n++;
      }
    }
  }
  
  return n==2;
}

/*** glutMouseFunc ***/
void mouse(int button, int state, int x, int y) {
  GLdouble model[16], proj[16];
  GLint view[4];
  if(button == GLUT_LEFT_BUTTON) {
    if(state == GLUT_DOWN) {
      rotatex0 = x; rotatey0 = y; // used by motion()
    }
    else {
      if(rotating) // finished dragging
        rotating = FALSE;
      else { // clicked, didn't drag
      
        /*** compute pick ray: ***/
        // store view matrices, including transformed modelview matrix:
        glMatrixMode(GL_MODELVIEW);
        glPushMatrix();
          displayMatrixTransform();
          glGetDoublev(GL_MODELVIEW_MATRIX, model);
        glPopMatrix();
        glGetDoublev(GL_PROJECTION_MATRIX, proj);
        glGetIntegerv(GL_VIEWPORT, view);
        // compute pick ray, which is line through near and far
        gluUnProject(x,view[3]-y,0, model,proj,view, near+0,near+1,near+2);
        gluUnProject(x,view[3]-y,1, model,proj,view, far+0,far+1,far+2);
        pickCube = segmentCubeIntersect(near, far, pick0, pick1);
        if(pickCube) {
          clickField(&field, pick0, pick1);
          fieldTexture();
        }
        glutPostRedisplay();
      }
    }
  }
}

/*** returns an angle in [0,360) equivalent to theta ***/
inline float anglef(float theta) {
  theta = fmodf(theta, 360);
  if(theta<0)
    return theta+360;
  return theta;
}

/*** glutMotionFunc ***/
void motion(int x, int y) {
  if(x!=rotatex0 || y!=rotatey0) { // dragging
    rotating = TRUE;
    thetax = anglef(thetax + omega*(y-rotatey0));
    thetay = anglef(thetay + omega*(x-rotatex0));
    rotatex0 = x; rotatey0 = y;
    glutPostRedisplay();
  }
}

/*** called at exit() ***/
void destroy(void) {
  free(tex2D);
  grid_free(&field);
}

int main(int argc, char **argv) {
  glutInit(&argc, argv);
  glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
  glutInitWindowSize(320, 240);
  glutCreateWindow("Fields");
  
  init();

  glutReshapeFunc(reshape);
  glutDisplayFunc(display);
  glutIdleFunc(idle);
  glutKeyboardFunc(keyboard);
  glutSpecialFunc(special);
  glutMouseFunc(mouse);
  glutMotionFunc(motion);
  
  atexit(destroy);

  glutMainLoop();
  return 0;
}
